"""
Test config option type enforcement
"""

import pytest

import salt.config


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], True),  # list
        ((1, 2, 3), True),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", False),  # str
        (True, False),  # bool
        (1, False),  # int
        (0.123, False),  # float
        (None, False),  # None
    ],
)
def test_list_types(option_value, expected):
    """
    List and tuple type config options return True when the value is a list. All
    other types return False
    modules_dirs is a list type config option
    """
    result = salt.config._validate_opts({"module_dirs": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", True),  # str
        (True, True),  # bool
        (1, True),  # int
        (0.123, True),  # float
        (None, True),  # None
    ],
)
def test_str_types(option_value, expected):
    """
    Str, bool, int, float, and none type config options return True when the
    value is a str. All other types return False
    user is a str type config option
    """
    result = salt.config._validate_opts({"user": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, True),  # dict
        ("str", False),  # str
        (True, False),  # bool
        (1, False),  # int
        (0.123, False),  # float
        (None, False),  # None
    ],
)
def test_dict_types(option_value, expected):
    """
    Dict type config options return True when the value is a dict. All other
    types return False
    file_roots is a dict type config option
    """
    result = salt.config._validate_opts({"file_roots": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", False),  # str
        (True, True),  # bool
        (1, False),  # int
        (0.123, False),  # float
        (None, False),  # None
    ],
)
def test_bool_types(option_value, expected):
    """
    Bool type config options return True when the value is a bool. All other
    types return False
    local is a bool type config option
    """
    result = salt.config._validate_opts({"local": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", False),  # str
        (True, False),  # bool
        (1, True),  # int
        (0.123, False),  # float
        (None, False),  # None
    ],
)
def test_int_types(option_value, expected):
    """
    Int type config options return True when the value is an int. All other
    types return False
    publish_port is an int type config option
    """
    result = salt.config._validate_opts({"publish_port": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", False),  # str
        (True, False),  # bool
        (1, True),  # int
        (0.123, True),  # float
        (None, False),  # None
    ],
)
def test_float_types(option_value, expected):
    """
    Float and int type config options return True when the value is a float. All
    other types return False
    ssh_timeout is a float type config option
    """
    result = salt.config._validate_opts({"ssh_timeout": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", True),  # str
        (True, True),  # bool
        (1, True),  # int
        (0.123, True),  # float
        (None, True),  # None
    ],
)
def test_none_str_types(option_value, expected):
    """
    Some config settings have two types, None and str. In that case str, bool,
    int, float, and None type options should evaluate as True. All others should
    return False.
    saltenv is a None, str type config option
    """
    result = salt.config._validate_opts({"saltenv": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", False),  # str
        (True, False),  # bool
        (1, True),  # int
        (0.123, False),  # float
        (None, True),  # None
    ],
)
def test_none_int_types(option_value, expected):
    """
    Some config settings have two types, None and int, which should evaluate as
    True. All others should return False.
    retry_dns_count is a None, int type config option
    """
    result = salt.config._validate_opts({"retry_dns_count": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", False),  # str
        (True, True),  # bool
        (1, False),  # int
        (0.123, False),  # float
        (None, True),  # None
    ],
)
def test_none_bool_types(option_value, expected):
    """
    Some config settings have two types, None and bool which should evaluate as
    True. All others should return False.
    ipv6 is a None, bool type config option
    """
    result = salt.config._validate_opts({"ipv6": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], True),  # list
        ((1, 2, 3), True),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", True),  # str
        (True, True),  # bool
        (1, True),  # int
        (0.123, True),  # float
        (None, True),  # None
    ],
)
def test_str_list_types(option_value, expected):
    """
    Some config settings have two types, str and list. In that case, list,
    tuple, str, bool, int, float, and None should evaluate as True. All others
    should return False.
    master is a str, list type config option
    """
    result = salt.config._validate_opts({"master": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", True),  # str
        (True, True),  # bool
        (1, True),  # int
        (0.123, True),  # float
        (None, True),  # None
    ],
)
def test_str_int_types(option_value, expected):
    """
    Some config settings have two types, str and int. In that case, str, bool,
    int, float, and None should evaluate as True. All others should return
    False.
    master_port is a str, int type config option
    """
    result = salt.config._validate_opts({"master_port": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, True),  # dict
        ("str", True),  # str
        (True, True),  # bool
        (1, True),  # int
        (0.123, True),  # float
        (None, True),  # None
    ],
)
def test_str_dict_types(option_value, expected):
    """
    Some config settings have two types, str and dict. In that case, dict, str,
    bool, int, float, and None should evaluate as True. All others should return
    False.
    id_function is a str, dict type config option
    """
    result = salt.config._validate_opts({"id_function": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], True),  # list
        ((1, 2, 3), True),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", True),  # str
        (True, True),  # bool
        (1, True),  # int
        (0.123, True),  # float
        (None, True),  # None
    ],
)
def test_str_tuple_types(option_value, expected):
    """
    Some config settings have two types, str and tuple. In that case, list,
    tuple, str, bool, int, float, and None should evaluate as True. All others
    should return False.
    log_fmt_logfile is a str, tuple type config option
    """
    result = salt.config._validate_opts({"log_fmt_logfile": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", True),  # str
        (True, True),  # bool
        (1, True),  # int
        (0.123, True),  # float
        (None, True),  # None
    ],
)
def test_str_bool_types(option_value, expected):
    """
    Some config settings have two types, str and bool. In that case, str, bool,
    int, float, and None should evaluate as True. All others should return
    False.
    update_url is a str, bool type config option
    """
    result = salt.config._validate_opts({"update_url": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, True),  # dict
        ("str", False),  # str
        (True, True),  # bool
        (1, False),  # int
        (0.123, False),  # float
        (None, False),  # None
    ],
)
def test_dict_bool_types(option_value, expected):
    """
    Some config settings have two types, dict and bool which should evaluate as
    True. All others should return False.
    token_expire_user_override is a dict, bool type config option
    """
    result = salt.config._validate_opts({"token_expire_user_override": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], True),  # list
        ((1, 2, 3), True),  # tuple
        ({"key": "value"}, True),  # dict
        ("str", False),  # str
        (True, False),  # bool
        (1, False),  # int
        (0.123, False),  # float
        (None, False),  # None
    ],
)
def test_dict_list_types(option_value, expected):
    """
    Some config settings have two types, dict and list. In that case, list,
    tuple, and dict should evaluate as True. All others should return False.
    nodegroups is a dict, list type config option
    """
    result = salt.config._validate_opts({"nodegroups": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, True),  # dict
        ("str", False),  # str
        (True, True),  # bool
        (1, False),  # int
        (0.123, False),  # float
        (None, True),  # None
    ],
)
def test_dict_bool_none_types(option_value, expected):
    """
    Some config settings have three types, dict, bool, and None which should
    evaluate as True. All others should return False.
    ssl is a dict, bool type config option
    """
    result = salt.config._validate_opts({"ssl": option_value})
    assert result is expected


@pytest.mark.parametrize(
    "option_value,expected",
    [
        ([1, 2, 3], False),  # list
        ((1, 2, 3), False),  # tuple
        ({"key": "value"}, False),  # dict
        ("str", False),  # str
        (True, True),  # bool
        (1, True),  # int
        (0.123, False),  # float
        (None, False),  # None
    ],
)
def test_bool_int_types(option_value, expected):
    """
    Some config settings have two types, bool and int. In that case, bool and
    int should evaluate as True. All others should return False.
    state_queue is a bool/int config option
    """
    result = salt.config._validate_opts({"state_queue": option_value})
    assert result is expected



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest
from tornado.httpclient import HTTPClient

import salt.modules.random_org as random_org
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase


def check_status():
    """
    Check the status of random.org
    """
    try:
        return HTTPClient().fetch("https://api.random.org/").code == 200
    except Exception:  # pylint: disable=broad-except
        return False


@pytest.mark.skip(
    reason="WAR ROOM 7/31/2019, test needs to allow for quotas of random website"
)
class RandomOrgTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.random_org
    """

    def setup_loader_modules(self):
        return {random_org: {}}

    def setUp(self):
        if check_status() is False:
            self.skipTest("External resource 'https://api.random.org/' not available")

    # 'getUsage' function tests: 1

    def test_getusage(self):
        """
        Test if it show current usages statistics.
        """
        ret = {"message": "No Random.org api key or api version found.", "res": False}
        self.assertDictEqual(random_org.getUsage(), ret)

        self.assertDictEqual(
            random_org.getUsage(api_key="peW", api_version="1"),
            {
                "bitsLeft": None,
                "requestsLeft": None,
                "res": True,
                "totalBits": None,
                "totalRequests": None,
            },
        )

    # 'generateIntegers' function tests: 1

    def test_generateintegers(self):
        """
        Test if it generate random integers.
        """
        ret1 = {"message": "No Random.org api key or api version found.", "res": False}
        self.assertDictEqual(random_org.generateIntegers(), ret1)

        ret2 = {"message": "Rquired argument, number is missing.", "res": False}
        self.assertDictEqual(
            random_org.generateIntegers(api_key="peW", api_version="1"), ret2
        )

        ret3 = {
            "message": "Number of integers must be between 1 and 10000",
            "res": False,
        }
        self.assertDictEqual(
            random_org.generateIntegers(
                api_key="peW", api_version="1", number="5", minimum="1", maximum="6"
            ),
            ret3,
        )

        ret4 = {
            "message": (
                "Minimum argument must be between -1,000,000,000 and 1,000,000,000"
            ),
            "res": False,
        }
        self.assertDictEqual(
            random_org.generateIntegers(
                api_key="peW", api_version="1", number=5, minimum="1", maximum="6"
            ),
            ret4,
        )

        ret5 = {
            "message": (
                "Maximum argument must be between -1,000,000,000 and 1,000,000,000"
            ),
            "res": False,
        }
        self.assertDictEqual(
            random_org.generateIntegers(
                api_key="peW", api_version="1", number=5, minimum=1, maximum="6"
            ),
            ret5,
        )

        ret6 = {"message": "Base must be either 2, 8, 10 or 16.", "res": False}
        self.assertDictEqual(
            random_org.generateIntegers(
                api_key="peW", api_version="1", number=5, minimum=1, maximum=6, base="2"
            ),
            ret6,
        )

        ret7 = {"message": "Parameter 'apiKey' is malformed", "res": False}
        self.assertDictEqual(
            random_org.generateIntegers(
                api_key="peW", api_version="1", number=5, minimum=1, maximum=6, base=2
            ),
            ret7,
        )

    # 'generateStrings' function tests: 1

    def test_generatestrings(self):
        """
        Test if it generate random strings.
        """
        ret1 = {"message": "No Random.org api key or api version found.", "res": False}
        self.assertDictEqual(random_org.generateStrings(), ret1)

        ret2 = {"message": "Required argument, number is missing.", "res": False}
        self.assertDictEqual(
            random_org.generateStrings(api_key="peW", api_version="1"), ret2
        )

        ret3 = {
            "message": "Number of strings must be between 1 and 10000",
            "res": False,
        }
        char = "abcdefghijklmnopqrstuvwxyz"
        self.assertDictEqual(
            random_org.generateStrings(
                api_key="peW", api_version="1", number="5", length="8", characters=char
            ),
            ret3,
        )

        ret3 = {"message": "Length of strings must be between 1 and 20", "res": False}
        self.assertDictEqual(
            random_org.generateStrings(
                api_key="peW", api_version="1", number=5, length="8", characters=char
            ),
            ret3,
        )

        ret3 = {"message": "Length of characters must be less than 80.", "res": False}
        self.assertDictEqual(
            random_org.generateStrings(
                api_key="peW", api_version="1", number=5, length=8, characters=char * 4
            ),
            ret3,
        )

        ret3 = {"message": "Parameter 'apiKey' is malformed", "res": False}
        self.assertDictEqual(
            random_org.generateStrings(
                api_key="peW", api_version="1", number=5, length=8, characters=char
            ),
            ret3,
        )

    # 'generateUUIDs' function tests: 1

    def test_generateuuids(self):
        """
        Test if it generate a list of random UUIDs.
        """
        ret1 = {"message": "No Random.org api key or api version found.", "res": False}
        self.assertDictEqual(random_org.generateUUIDs(), ret1)

        ret2 = {"message": "Required argument, number is missing.", "res": False}
        self.assertDictEqual(
            random_org.generateUUIDs(api_key="peW", api_version="1"), ret2
        )

        ret3 = {"message": "Number of UUIDs must be between 1 and 1000", "res": False}
        self.assertDictEqual(
            random_org.generateUUIDs(api_key="peW", api_version="1", number="5"), ret3
        )

        ret3 = {"message": "Parameter 'apiKey' is malformed", "res": False}
        self.assertDictEqual(
            random_org.generateUUIDs(api_key="peW", api_version="1", number=5), ret3
        )

    # 'generateDecimalFractions' function tests: 1

    @pytest.mark.flaky(max_runs=4)
    def test_generatedecimalfractions(self):
        """
        Test if it generates true random decimal fractions.
        """
        ret1 = {"message": "No Random.org api key or api version found.", "res": False}
        self.assertDictEqual(random_org.generateDecimalFractions(), ret1)

        ret2 = {"message": "Required argument, number is missing.", "res": False}
        self.assertDictEqual(
            random_org.generateDecimalFractions(api_key="peW", api_version="1"), ret2
        )

        ret3 = {
            "message": "Number of decimal fractions must be between 1 and 10000",
            "res": False,
        }
        self.assertDictEqual(
            random_org.generateDecimalFractions(
                api_key="peW",
                api_version="1",
                number="5",
                decimalPlaces="4",
                replacement=True,
            ),
            ret3,
        )

        ret4 = {
            "message": "Number of decimal places must be between 1 and 20",
            "res": False,
        }
        self.assertDictEqual(
            random_org.generateDecimalFractions(
                api_key="peW",
                api_version="1",
                number=5,
                decimalPlaces="4",
                replacement=True,
            ),
            ret4,
        )

        ret5 = {"message": "Parameter 'apiKey' is malformed", "res": False}
        self.assertDictEqual(
            random_org.generateDecimalFractions(
                api_key="peW",
                api_version="1",
                number=5,
                decimalPlaces=4,
                replacement=True,
            ),
            ret5,
        )

    # 'generateGaussians' function tests: 1

    @pytest.mark.flaky(max_runs=4)
    def test_generategaussians(self):
        """
        Test if it generates true random numbers from a
        Gaussian distribution (also known as a normal distribution).
        """
        ret1 = {"message": "No Random.org api key or api version found.", "res": False}
        self.assertDictEqual(random_org.generateGaussians(), ret1)

        ret2 = {"message": "Required argument, number is missing.", "res": False}
        self.assertDictEqual(
            random_org.generateGaussians(api_key="peW", api_version="1"), ret2
        )

        ret3 = {
            "message": "Number of decimal fractions must be between 1 and 10000",
            "res": False,
        }
        self.assertDictEqual(
            random_org.generateGaussians(
                api_key="peW",
                api_version="1",
                number="5",
                mean="0.0",
                standardDeviation="1.0",
                significantDigits="8",
            ),
            ret3,
        )

        ret4 = {
            "message": "The distribution's mean must be between -1000000 and 1000000",
            "res": False,
        }
        self.assertDictEqual(
            random_org.generateGaussians(
                api_key="peW",
                api_version="1",
                number=5,
                mean="0.0",
                standardDeviation="1.0",
                significantDigits="8",
            ),
            ret4,
        )

        ret5 = {
            "message": (
                "The distribution's standard deviation must be"
                " between -1000000 and 1000000"
            ),
            "res": False,
        }
        self.assertDictEqual(
            random_org.generateGaussians(
                api_key="peW",
                api_version="1",
                number=5,
                mean=0.0,
                standardDeviation="1.0",
                significantDigits="8",
            ),
            ret5,
        )

        ret6 = {
            "message": "The number of significant digits must be between 2 and 20",
            "res": False,
        }
        self.assertDictEqual(
            random_org.generateGaussians(
                api_key="peW",
                api_version="1",
                number=5,
                mean=0.0,
                standardDeviation=1.0,
                significantDigits="8",
            ),
            ret6,
        )

        ret7 = {"message": "Parameter 'apiKey' is malformed", "res": False}
        self.assertDictEqual(
            random_org.generateGaussians(
                api_key="peW",
                api_version="1",
                number=5,
                mean=0.0,
                standardDeviation=1.0,
                significantDigits=8,
            ),
            ret7,
        )

    # 'generateBlobs' function tests: 1

    def test_generateblobs(self):
        """
        Test if it list all Slack users.
        """
        ret1 = {"message": "No Random.org api key or api version found.", "res": False}
        self.assertDictEqual(random_org.generateBlobs(), ret1)

        ret2 = {"message": "Required argument, number is missing.", "res": False}
        self.assertDictEqual(
            random_org.generateBlobs(api_key="peW", api_version="1"), ret2
        )

        ret3 = {"message": "Number of blobs must be between 1 and 100", "res": False}
        self.assertDictEqual(
            random_org.generateBlobs(
                api_key="peW", api_version="1", number="5", size="1"
            ),
            ret3,
        )

        ret4 = {"message": "Number of blobs must be between 1 and 100", "res": False}
        self.assertDictEqual(
            random_org.generateBlobs(api_key="peW", api_version="1", number=5, size=1),
            ret4,
        )

        ret5 = {"message": "Format must be either base64 or hex.", "res": False}
        self.assertDictEqual(
            random_org.generateBlobs(
                api_key="peW", api_version="1", number=5, size=8, format="oct"
            ),
            ret5,
        )

        ret6 = {"message": "Parameter 'apiKey' is malformed", "res": False}
        self.assertDictEqual(
            random_org.generateBlobs(
                api_key="peW", api_version="1", number=5, size=8, format="hex"
            ),
            ret6,
        )



"""
    :codeauthor: :email:`Rod McKenzie (roderick.mckenzie@morganstanley.com)`
    :codeauthor: :email:`Alexandru Bleotu (alexandru.bleotu@morganstanley.com)`

    salt.config.schemas.vcenter
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    VCenter configuration schemas
"""

from salt.utils.schema import ArrayItem, BooleanItem, IntegerItem, Schema, StringItem


class VCenterEntitySchema(Schema):
    """
    Entity Schema for a VCenter.
    """

    title = "VCenter Entity Schema"
    description = "VCenter entity schema"
    type = StringItem(
        title="Type",
        description="Specifies the entity type",
        required=True,
        enum=["vcenter"],
    )

    vcenter = StringItem(
        title="vCenter", description="Specifies the vcenter hostname", required=True
    )


class VCenterProxySchema(Schema):
    """
    Schema for the configuration for the proxy to connect to a VCenter.
    """

    title = "VCenter Proxy Connection Schema"
    description = "Schema that describes the connection to a VCenter"
    additional_properties = False
    proxytype = StringItem(required=True, enum=["vcenter"])
    vcenter = StringItem(required=True, pattern=r"[^\s]+")
    mechanism = StringItem(required=True, enum=["userpass", "sspi"])
    username = StringItem()
    passwords = ArrayItem(min_items=1, items=StringItem(), unique_items=True)
    verify_ssl = BooleanItem()
    ca_bundle = StringItem()

    domain = StringItem()
    principal = StringItem(default="host")
    protocol = StringItem(default="https")
    port = IntegerItem(minimum=1)



import pytest

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module", autouse=True)
def pillar_tree(base_env_pillar_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - basic
      '127.0.0.1':
        - basic
    """
    basic_pillar_file = """
    monty: python
    knights:
      - Lancelot
      - Galahad
      - Bedevere
      - Robin
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    basic_tempfile = pytest.helpers.temp_file(
        "basic.sls", basic_pillar_file, base_env_pillar_tree_root_dir
    )

    with top_tempfile, basic_tempfile:
        yield


@pytest.fixture()
def pillar_filter_by_lookup():
    return {
        "common": {
            "has_common": True,
        },
        "custom_default": {
            "defaulted": True,
        },
        "merge": {
            "merged": True,
        },
    }


def test_pillar_items(salt_ssh_cli):
    """
    test pillar.items with salt-ssh
    """
    ret = salt_ssh_cli.run("pillar.items")
    assert ret.returncode == 0
    assert ret.data
    pillar_items = ret.data
    assert "monty" in pillar_items
    assert pillar_items["monty"] == "python"
    assert "knights" in pillar_items
    assert pillar_items["knights"] == ["Lancelot", "Galahad", "Bedevere", "Robin"]


def test_pillar_get(salt_ssh_cli):
    """
    test pillar.get with salt-ssh
    """
    ret = salt_ssh_cli.run("pillar.get", "monty")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == "python"


def test_pillar_get_doesnotexist(salt_ssh_cli):
    """
    test pillar.get when pillar does not exist with salt-ssh
    """
    ret = salt_ssh_cli.run("pillar.get", "doesnotexist")
    assert ret.returncode == 0
    assert ret.data == ""


def test_pillar_filter_by(salt_ssh_cli, pillar_filter_by_lookup):
    """
    test pillar.filter_by with salt-ssh
    """
    pillar_filter_by_lookup["python"] = {"filtered": True}
    ret = salt_ssh_cli.run(
        "pillar.filter_by",
        pillar_filter_by_lookup,
        pillar="monty",
        merge=pillar_filter_by_lookup["merge"],
        base="common",
        default="custom_default",
    )
    assert ret.returncode == 0
    assert ret.data
    assert "has_common" in ret.data
    assert "filtered" in ret.data
    assert "merged" in ret.data
    assert "defaulted" not in ret.data


def test_pillar_filter_by_default(salt_ssh_cli, pillar_filter_by_lookup):
    """
    test pillar.filter_by default param with salt-ssh
    """
    ret = salt_ssh_cli.run(
        "pillar.filter_by",
        pillar_filter_by_lookup,
        pillar="monty",
        merge=pillar_filter_by_lookup["merge"],
        base="common",
        default="custom_default",
    )
    assert ret.returncode == 0
    assert ret.data
    assert "has_common" in ret.data
    assert "merged" in ret.data
    assert "defaulted" in ret.data



import pytest

pytestmark = [
    pytest.mark.skip_on_windows,
]


def test_salt_key(salt_key_cli, salt_minion):
    """
    Test running salt-key -L
    """
    ret = salt_key_cli.run("-L")
    assert ret.data
    assert salt_minion.id in ret.data["minions"]



import pytest


@pytest.mark.parametrize(
    "env,warn_expected",
    [
        ({"PYTHONWARNINGS": "ignore"}, False),
        ({"PYTHONWARNINGS": "test"}, True),
        ({}, True),
    ],
)
def test_deprecation_warnings(salt_call_cli, env, warn_expected):
    """
    Test to ensure when env variable PYTHONWARNINGS=ignore
    is set that we do not add warning to output.
    And when it's not set to ignore the warning will show.
    """
    ret = salt_call_cli.run("--local", "test.deprecation_warning", env=env)
    if warn_expected:
        assert "DeprecationWarning" in ret.stderr
        assert ret.stderr.count("DeprecationWarning") >= 2
    else:
        assert "DeprecationWarning" not in ret.stderr
        assert ret.data


import salt.modules.test as test


def test_raise_exception():
    """
    Add test raising an exception in test module.
    """
    msg = "message"
    with pytest.raises(Exception) as err:
        test.exception(message=msg)
    assert err.match(msg)



"""
Integration tests for the ini_manage state
"""

import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


def test_options_present(salt_master, salt_call_cli):
    """
    test ini.options_present when the file
    does not exist and then run it again
    when it does exist and run it again when
    we want to add more sections to the ini
    """
    with pytest.helpers.temp_file("ini_file.ini") as tpath:
        content = """
        test_ini:
          ini.options_present:
            - name: {}
            - sections:
                general:
                  server_hostname: foo.com
                  server_port: 1234
        """.format(
            tpath
        )

        with salt_master.state_tree.base.temp_file("manage_ini.sls", content):
            ret = salt_call_cli.run("--local", "state.apply", "manage_ini")
            assert ret.data[next(iter(ret.data))]["changes"] == {
                "general": {
                    "before": None,
                    "after": {"server_hostname": "foo.com", "server_port": "1234"},
                }
            }

        content = """
        test_ini:
          ini.options_present:
            - name: {}
            - sections:
                general:
                  server_hostname: foo.com
                  server_port: 1234
                  server_user: saltfoo
        """.format(
            tpath
        )

        with salt_master.state_tree.base.temp_file("manage_ini.sls", content):
            # check to see adding a new section works
            ret = salt_call_cli.run("--local", "state.apply", "manage_ini")
            assert ret.data[next(iter(ret.data))]["changes"] == {
                "general": {"server_user": {"before": None, "after": "saltfoo"}}
            }

            # check when no changes are expected
            ret = salt_call_cli.run("--local", "state.apply", "manage_ini")
            assert ret.data[next(iter(ret.data))]["changes"] == {}



"""
Provide the service module for the dummy proxy used in integration tests
"""

import logging

import salt.utils.platform

log = logging.getLogger(__name__)

__func_alias__ = {"list_": "list"}


# Define the module's virtual name
__virtualname__ = "service"


def __virtual__():
    """
    Only work on systems that are a proxy minion
    """
    try:
        if salt.utils.platform.is_proxy() and __opts__["proxy"]["proxytype"] == "dummy":
            return __virtualname__
    except KeyError:
        return (
            False,
            "The dummyproxy_service execution module failed to load. Check "
            "the proxy key in pillar or /etc/salt/proxy.",
        )

    return (
        False,
        "The dummyproxy_service execution module failed to load: only works "
        "on the integration testsuite dummy proxy minion.",
    )


def get_all():
    """
    Return a list of all available services

    .. versionadded:: 2016.11.3

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_all
    """
    proxy_fn = "dummy.service_list"
    return __proxy__[proxy_fn]()


def list_():
    """
    Return a list of all available services.

    .. versionadded:: 2016.11.3

    CLI Example:

    .. code-block:: bash

        salt '*' service.list
    """
    return get_all()


def start(name, sig=None):
    """
    Start the specified service on the dummy

    .. versionadded:: 2016.11.3

    CLI Example:

    .. code-block:: bash

        salt '*' service.start <service name>
    """

    proxy_fn = "dummy.service_start"
    return __proxy__[proxy_fn](name)


def stop(name, sig=None):
    """
    Stop the specified service on the dummy

    .. versionadded:: 2016.11.3

    CLI Example:

    .. code-block:: bash

        salt '*' service.stop <service name>
    """
    proxy_fn = "dummy.service_stop"
    return __proxy__[proxy_fn](name)


def restart(name, sig=None):
    """
    Restart the specified service with dummy.

    .. versionadded:: 2016.11.3

    CLI Example:

    .. code-block:: bash

        salt '*' service.restart <service name>
    """

    proxy_fn = "dummy.service_restart"
    return __proxy__[proxy_fn](name)


def status(name, sig=None):
    """
    Return the status for a service via dummy, returns a bool
    whether the service is running.

    .. versionadded:: 2016.11.3

    CLI Example:

    .. code-block:: bash

        salt '*' service.status <service name>
    """

    proxy_fn = "dummy.service_status"
    resp = __proxy__[proxy_fn](name)
    if resp["comment"] == "stopped":
        return False
    if resp["comment"] == "running":
        return True


def running(name, sig=None):
    """
    Return whether this service is running.

    .. versionadded:: 2016.11.3

    """
    return status(name).get(name, False)


def enabled(name, sig=None):
    """
    Only the 'redbull' service is 'enabled' in the test

    .. versionadded:: 2016.11.3

    """
    return name == "redbull"



import logging
import random
import string

import pytest

import salt.config
import salt.loader
import salt.states.boto_lambda as boto_lambda
import salt.utils.json
from tests.support.mock import MagicMock, patch

# pylint: disable=import-error,no-name-in-module
from tests.unit.modules.test_boto_lambda import TempZipFile

boto = pytest.importorskip("boto")
boto3 = pytest.importorskip("boto3", "1.2.1")
botocore = pytest.importorskip("botocore", "1.5.2")

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.slow_test,
]


class GlobalConfig:
    region = "us-east-1"
    access_key = "GKTADJGHEIQSXMKKRBJ08H"
    secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
    conn_parameters = {
        "region": region,
        "key": access_key,
        "keyid": secret_key,
        "profile": {},
    }
    error_message = (
        "An error occurred (101) when calling the {0} operation: Test-defined error"
    )
    error_content = {"Error": {"Code": 101, "Message": "Test-defined error"}}
    function_ret = dict(
        FunctionName="testfunction",
        Runtime="python2.7",
        Role="arn:aws:iam::1234:role/functionrole",
        Handler="handler",
        Description="abcdefg",
        Timeout=5,
        MemorySize=128,
        CodeSha256="abcdef",
        CodeSize=199,
        FunctionArn="arn:lambda:us-east-1:1234:Something",
        LastModified="yes",
        VpcConfig={"SubnetIds": [], "SecurityGroupIds": []},
    )
    alias_ret = dict(
        AliasArn="arn:lambda:us-east-1:1234:Something",
        Name="testalias",
        FunctionVersion="3",
        Description="Alias description",
    )
    event_source_mapping_ret = dict(
        UUID="1234-1-123",
        BatchSize=123,
        EventSourceArn="arn:aws:dynamodb:us-east-1:1234::Something",
        FunctionArn="arn:aws:lambda:us-east-1:1234:function:myfunc",
        LastModified="yes",
        LastProcessingResult="SUCCESS",
        State="Enabled",
        StateTransitionReason="Random",
    )


@pytest.fixture
def global_config():
    GlobalConfig.conn_parameters["key"] = "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
    )
    params = GlobalConfig()
    return params


@pytest.fixture
def session_instance():
    with patch("boto3.session.Session") as patched_session:
        yield patched_session()


@pytest.fixture
def configure_loader_modules(minion_opts):
    minion_opts["grains"] = salt.loader.grains(minion_opts)
    ctx = {}
    utils = salt.loader.utils(
        minion_opts,
        whitelist=["boto", "boto3", "args", "systemd", "path", "platform", "reg"],
        context=ctx,
    )
    serializers = salt.loader.serializers(minion_opts)
    funcs = salt.loader.minion_mods(
        minion_opts, context=ctx, utils=utils, whitelist=["boto_lambda"]
    )
    salt_states = salt.loader.states(
        opts=minion_opts,
        functions=funcs,
        utils=utils,
        whitelist=["boto_lambda"],
        serializers=serializers,
    )
    return {
        boto_lambda: {
            "__opts__": minion_opts,
            "__salt__": funcs,
            "__utils__": utils,
            "__states__": salt_states,
            "__serializers__": serializers,
        }
    }


@pytest.mark.slow_test
def test_present_when_function_does_not_exist_func(global_config, session_instance):
    """
    Tests present on a function that does not exist.
    """
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_functions.side_effect = [
        {"Functions": []},
        {"Functions": [global_config.function_ret]},
    ]
    conn.create_function.return_value = global_config.function_ret
    with patch.dict(
        boto_lambda.__salt__,
        {"boto_iam.get_account_id": MagicMock(return_value="1234")},
    ):
        with TempZipFile() as zipfile:
            result = boto_lambda.__states__["boto_lambda.function_present"](
                "function present",
                FunctionName=global_config.function_ret["FunctionName"],
                Runtime=global_config.function_ret["Runtime"],
                Role=global_config.function_ret["Role"],
                Handler=global_config.function_ret["Handler"],
                ZipFile=zipfile,
            )

    assert result["result"]
    assert (
        result["changes"]["new"]["function"]["FunctionName"]
        == global_config.function_ret["FunctionName"]
    )


@pytest.mark.slow_test
def test_present_when_function_exists_func(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_functions.return_value = {"Functions": [global_config.function_ret]}
    conn.update_function_code.return_value = global_config.function_ret

    with patch.dict(
        boto_lambda.__salt__,
        {"boto_iam.get_account_id": MagicMock(return_value="1234")},
    ):
        with TempZipFile() as zipfile:
            with patch("hashlib.sha256") as sha256:
                with patch("os.path.getsize", return_value=199):
                    sha = sha256()
                    digest = sha.digest()
                    encoded = sha.encode()
                    encoded.strip.return_value = global_config.function_ret[
                        "CodeSha256"
                    ]
                    result = boto_lambda.__states__["boto_lambda.function_present"](
                        "function present",
                        FunctionName=global_config.function_ret["FunctionName"],
                        Runtime=global_config.function_ret["Runtime"],
                        Role=global_config.function_ret["Role"],
                        Handler=global_config.function_ret["Handler"],
                        ZipFile=zipfile,
                        Description=global_config.function_ret["Description"],
                        Timeout=global_config.function_ret["Timeout"],
                    )
    assert result["result"]
    assert result["changes"] == {}


@pytest.mark.slow_test
def test_present_with_failure_func(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_functions.side_effect = [
        {"Functions": []},
        {"Functions": [global_config.function_ret]},
    ]
    conn.create_function.side_effect = botocore.exceptions.ClientError(
        global_config.error_content, "create_function"
    )
    with patch.dict(
        boto_lambda.__salt__,
        {"boto_iam.get_account_id": MagicMock(return_value="1234")},
    ):
        with TempZipFile() as zipfile:
            with patch("hashlib.sha256") as sha256:
                with patch("os.path.getsize", return_value=199):
                    sha = sha256()
                    digest = sha.digest()
                    encoded = sha.encode()
                    encoded.strip.return_value = global_config.function_ret[
                        "CodeSha256"
                    ]
                    result = boto_lambda.__states__["boto_lambda.function_present"](
                        "function present",
                        FunctionName=global_config.function_ret["FunctionName"],
                        Runtime=global_config.function_ret["Runtime"],
                        Role=global_config.function_ret["Role"],
                        Handler=global_config.function_ret["Handler"],
                        ZipFile=zipfile,
                        Description=global_config.function_ret["Description"],
                        Timeout=global_config.function_ret["Timeout"],
                    )
    assert not result["result"]
    assert "An error occurred" in result["comment"]


def test_absent_when_function_does_not_exist_func(global_config, session_instance):
    """
    Tests absent on a function that does not exist.
    """
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_functions.return_value = {"Functions": [global_config.function_ret]}
    result = boto_lambda.__states__["boto_lambda.function_absent"]("test", "myfunc")
    assert result["result"]
    assert result["changes"] == {}


def test_absent_when_function_exists_func(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_functions.return_value = {"Functions": [global_config.function_ret]}
    result = boto_lambda.__states__["boto_lambda.function_absent"](
        "test", global_config.function_ret["FunctionName"]
    )
    assert result["result"]
    assert result["changes"]["new"]["function"] is None


def test_absent_with_failure_func(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_functions.return_value = {"Functions": [global_config.function_ret]}
    conn.delete_function.side_effect = botocore.exceptions.ClientError(
        global_config.error_content, "delete_function"
    )
    result = boto_lambda.__states__["boto_lambda.function_absent"](
        "test", global_config.function_ret["FunctionName"]
    )
    assert not result["result"]
    assert "An error occurred" in result["comment"]


@pytest.mark.slow_test
def test_present_when_function_exists_and_permissions_func(
    global_config, session_instance
):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_functions.return_value = {"Functions": [global_config.function_ret]}
    conn.update_function_code.return_value = global_config.function_ret
    conn.get_policy.return_value = {
        "Policy": salt.utils.json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Condition": {
                            "ArnLike": {
                                "AWS:SourceArn": (
                                    "arn:aws:events:us-east-1:9999999999:rule/fooo"
                                )
                            }
                        },
                        "Action": "lambda:InvokeFunction",
                        "Resource": "arn:aws:lambda:us-east-1:999999999999:function:testfunction",
                        "Effect": "Allow",
                        "Principal": {"Service": "events.amazonaws.com"},
                        "Sid": "AWSEvents_foo-bar999999999999",
                    }
                ],
                "Id": "default",
            }
        )
    }

    with patch.dict(
        boto_lambda.__salt__,
        {"boto_iam.get_account_id": MagicMock(return_value="1234")},
    ):
        with TempZipFile() as zipfile:
            with patch("hashlib.sha256") as sha256:
                with patch("os.path.getsize", return_value=199):
                    sha = sha256()
                    digest = sha.digest()
                    encoded = sha.encode()
                    encoded.strip.return_value = global_config.function_ret[
                        "CodeSha256"
                    ]
                    result = boto_lambda.__states__["boto_lambda.function_present"](
                        "function present",
                        FunctionName=global_config.function_ret["FunctionName"],
                        Runtime=global_config.function_ret["Runtime"],
                        Role=global_config.function_ret["Role"],
                        Handler=global_config.function_ret["Handler"],
                        ZipFile=zipfile,
                        Description=global_config.function_ret["Description"],
                        Timeout=global_config.function_ret["Timeout"],
                    )
    assert result["result"]
    assert result["changes"] == {
        "old": {
            "Permissions": {
                "AWSEvents_foo-bar999999999999": {
                    "Action": "lambda:InvokeFunction",
                    "Principal": "events.amazonaws.com",
                    "SourceArn": "arn:aws:events:us-east-1:9999999999:rule/fooo",
                }
            }
        },
        "new": {"Permissions": {"AWSEvents_foo-bar999999999999": {}}},
    }


def test_present_when_alias_does_not_exist(global_config, session_instance):
    """
    Tests present on a alias that does not exist.
    """
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_aliases.side_effect = [
        {"Aliases": []},
        {"Aliases": [global_config.alias_ret]},
    ]
    conn.create_alias.return_value = global_config.alias_ret
    result = boto_lambda.__states__["boto_lambda.alias_present"](
        "alias present",
        FunctionName="testfunc",
        Name=global_config.alias_ret["Name"],
        FunctionVersion=global_config.alias_ret["FunctionVersion"],
    )

    assert result["result"]
    assert result["changes"]["new"]["alias"]["Name"] == global_config.alias_ret["Name"]


def test_present_when_alias_exists(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_aliases.return_value = {"Aliases": [global_config.alias_ret]}
    conn.create_alias.return_value = global_config.alias_ret
    result = boto_lambda.__states__["boto_lambda.alias_present"](
        "alias present",
        FunctionName="testfunc",
        Name=global_config.alias_ret["Name"],
        FunctionVersion=global_config.alias_ret["FunctionVersion"],
        Description=global_config.alias_ret["Description"],
    )
    assert result["result"]
    assert result["changes"] == {}


@pytest.mark.slow_test
def test_present_with_failure_glob(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_aliases.side_effect = [
        {"Aliases": []},
        {"Aliases": [global_config.alias_ret]},
    ]
    conn.create_alias.side_effect = botocore.exceptions.ClientError(
        global_config.error_content, "create_alias"
    )
    result = boto_lambda.__states__["boto_lambda.alias_present"](
        "alias present",
        FunctionName="testfunc",
        Name=global_config.alias_ret["Name"],
        FunctionVersion=global_config.alias_ret["FunctionVersion"],
    )
    assert not result["result"]
    assert "An error occurred" in result["comment"]


def test_absent_when_alias_does_not_exist(global_config, session_instance):
    """
    Tests absent on a alias that does not exist.
    """
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_aliases.return_value = {"Aliases": [global_config.alias_ret]}
    result = boto_lambda.__states__["boto_lambda.alias_absent"](
        "alias absent", FunctionName="testfunc", Name="myalias"
    )
    assert result["result"]
    assert result["changes"] == {}


def test_absent_when_alias_exists(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_aliases.return_value = {"Aliases": [global_config.alias_ret]}
    result = boto_lambda.__states__["boto_lambda.alias_absent"](
        "alias absent", FunctionName="testfunc", Name=global_config.alias_ret["Name"]
    )
    assert result["result"]
    assert result["changes"]["new"]["alias"] is None


def test_absent_with_failure(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_aliases.return_value = {"Aliases": [global_config.alias_ret]}
    conn.delete_alias.side_effect = botocore.exceptions.ClientError(
        global_config.error_content, "delete_alias"
    )
    result = boto_lambda.__states__["boto_lambda.alias_absent"](
        "alias absent", FunctionName="testfunc", Name=global_config.alias_ret["Name"]
    )
    assert not result["result"]
    assert "An error occurred" in result["comment"]


def test_present_when_event_source_mapping_does_not_exist(
    global_config, session_instance
):
    """
    Tests present on a event_source_mapping that does not exist.
    """
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_event_source_mappings.side_effect = [
        {"EventSourceMappings": []},
        {"EventSourceMappings": [global_config.event_source_mapping_ret]},
    ]
    conn.get_event_source_mapping.return_value = global_config.event_source_mapping_ret
    conn.create_event_source_mapping.return_value = (
        global_config.event_source_mapping_ret
    )
    result = boto_lambda.__states__["boto_lambda.event_source_mapping_present"](
        "event source mapping present",
        EventSourceArn=global_config.event_source_mapping_ret["EventSourceArn"],
        FunctionName="myfunc",
        StartingPosition="LATEST",
    )

    assert result["result"]
    assert (
        result["changes"]["new"]["event_source_mapping"]["UUID"]
        == global_config.event_source_mapping_ret["UUID"]
    )


def test_present_when_event_source_mapping_exists(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_event_source_mappings.return_value = {
        "EventSourceMappings": [global_config.event_source_mapping_ret]
    }
    conn.get_event_source_mapping.return_value = global_config.event_source_mapping_ret
    conn.create_event_source_mapping.return_value = (
        global_config.event_source_mapping_ret
    )
    result = boto_lambda.__states__["boto_lambda.event_source_mapping_present"](
        "event source mapping present",
        EventSourceArn=global_config.event_source_mapping_ret["EventSourceArn"],
        FunctionName=global_config.event_source_mapping_ret["FunctionArn"],
        StartingPosition="LATEST",
        BatchSize=global_config.event_source_mapping_ret["BatchSize"],
    )
    assert result["result"]
    assert result["changes"] == {}


@pytest.mark.slow_test
def test_present_with_failure(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_event_source_mappings.side_effect = [
        {"EventSourceMappings": []},
        {"EventSourceMappings": [global_config.event_source_mapping_ret]},
    ]
    conn.get_event_source_mapping.return_value = global_config.event_source_mapping_ret
    conn.create_event_source_mapping.side_effect = botocore.exceptions.ClientError(
        global_config.error_content, "create_event_source_mapping"
    )
    result = boto_lambda.__states__["boto_lambda.event_source_mapping_present"](
        "event source mapping present",
        EventSourceArn=global_config.event_source_mapping_ret["EventSourceArn"],
        FunctionName=global_config.event_source_mapping_ret["FunctionArn"],
        StartingPosition="LATEST",
        BatchSize=global_config.event_source_mapping_ret["BatchSize"],
    )
    assert not result["result"]
    assert "An error occurred" in result["comment"]


def test_absent_when_event_source_mapping_does_not_exist(
    global_config, session_instance
):
    """
    Tests absent on a event_source_mapping that does not exist.
    """
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_event_source_mappings.return_value = {"EventSourceMappings": []}
    result = boto_lambda.__states__["boto_lambda.event_source_mapping_absent"](
        "event source mapping absent",
        EventSourceArn=global_config.event_source_mapping_ret["EventSourceArn"],
        FunctionName="myfunc",
    )
    assert result["result"]
    assert result["changes"] == {}


def test_absent_when_event_source_mapping_exists(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_event_source_mappings.return_value = {
        "EventSourceMappings": [global_config.event_source_mapping_ret]
    }
    conn.get_event_source_mapping.return_value = global_config.event_source_mapping_ret
    result = boto_lambda.__states__["boto_lambda.event_source_mapping_absent"](
        "event source mapping absent",
        EventSourceArn=global_config.event_source_mapping_ret["EventSourceArn"],
        FunctionName="myfunc",
    )
    assert result["result"]
    assert result["changes"]["new"]["event_source_mapping"] is None


def test_absent_with_failure_glob(global_config, session_instance):
    conn = MagicMock()
    session_instance.client.return_value = conn

    conn.list_event_source_mappings.return_value = {
        "EventSourceMappings": [global_config.event_source_mapping_ret]
    }
    conn.get_event_source_mapping.return_value = global_config.event_source_mapping_ret
    conn.delete_event_source_mapping.side_effect = botocore.exceptions.ClientError(
        global_config.error_content, "delete_event_source_mapping"
    )
    result = boto_lambda.__states__["boto_lambda.event_source_mapping_absent"](
        "event source mapping absent",
        EventSourceArn=global_config.event_source_mapping_ret["EventSourceArn"],
        FunctionName="myfunc",
    )
    assert not result["result"]
    assert "An error occurred" in result["comment"]



import pytest

import salt.utils.vault as vaultutil
import salt.utils.vault.api as vapi
import salt.utils.vault.client as vclient
from tests.support.mock import Mock, patch


@pytest.fixture
def entity_lookup_response():
    return {
        "data": {
            "aliases": [],
            "creation_time": "2017-11-13T21:01:33.543497Z",
            "direct_group_ids": [],
            "group_ids": [],
            "id": "043fedec-967d-b2c9-d3af-0c467b04e1fd",
            "inherited_group_ids": [],
            "last_update_time": "2017-11-13T21:01:33.543497Z",
            "merged_entity_ids": None,
            "metadata": None,
            "name": "test-minion",
            "policies": None,
        }
    }


@pytest.fixture
def entity_fetch_response():
    return {
        "data": {
            "aliases": [],
            "creation_time": "2018-09-19T17:20:27.705389973Z",
            "direct_group_ids": [],
            "disabled": False,
            "group_ids": [],
            "id": "test-entity-id",
            "inherited_group_ids": [],
            "last_update_time": "2018-09-19T17:20:27.705389973Z",
            "merged_entity_ids": None,
            "metadata": {
                "minion-id": "test-minion",
            },
            "name": "salt_minion_test-minion",
            "policies": [
                "default",
                "saltstack/minions",
                "saltstack/minion/test-minion",
            ],
        }
    }


@pytest.fixture
def secret_id_response():
    return {
        "request_id": "0e8c388e-2cb6-bcb2-83b7-625127d568bb",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": {
            "secret_id_accessor": "84896a0c-1347-aa90-a4f6-aca8b7558780",
            "secret_id": "841771dc-11c9-bbc7-bcac-6a3945a69cd9",
            "secret_id_ttl": 60,
        },
    }


@pytest.fixture
def secret_id_lookup_accessor_response():
    return {
        "request_id": "28f2f9fb-26c0-6022-4970-baeb6366b085",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": {
            "cidr_list": [],
            "creation_time": "2022-09-09T15:11:28.358490481+00:00",
            "expiration_time": "2022-10-11T15:11:28.358490481+00:00",
            "last_updated_time": "2022-09-09T15:11:28.358490481+00:00",
            "metadata": {},
            "secret_id_accessor": "0380eb9f-3041-1c1c-234c-fde31a1a1fc1",
            "secret_id_num_uses": 1,
            "secret_id_ttl": 9999999999,
            "token_bound_cidrs": [],
        },
        "warnings": None,
    }


@pytest.fixture
def wrapped_response():
    return {
        "request_id": "",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": None,
        "warnings": None,
        "wrap_info": {
            "token": "test-wrapping-token",
            "accessor": "test-wrapping-token-accessor",
            "ttl": 180,
            "creation_time": "2022-09-10T13:37:12.123456789+00:00",
            "creation_path": "whatever/not/checked/here",
            "wrapped_accessor": "84896a0c-1347-aa90-a4f6-aca8b7558780",
        },
    }


@pytest.fixture
def approle_meta(secret_id_serialized):
    return {
        "bind_secret_id": True,
        "local_secret_ids": False,
        "secret_id_bound_cidrs": [],
        "secret_id_num_uses": secret_id_serialized["secret_id_num_uses"],
        "secret_id_ttl": secret_id_serialized["secret_id_ttl"],
        "token_bound_cidrs": [],
        "token_explicit_max_ttl": 9999999999,
        "token_max_ttl": 0,
        "token_no_default_policy": False,
        "token_num_uses": 1,
        "token_period": 0,
        "token_policies": ["default"],
        "token_ttl": 0,
        "token_type": "default",
    }


@pytest.fixture
def secret_id_serialized(secret_id_response):
    return {
        "secret_id": secret_id_response["data"]["secret_id"],
        "secret_id_ttl": secret_id_response["data"]["secret_id_ttl"],
        "secret_id_num_uses": 1,
        # + creation_time
        # + expire_time
    }


@pytest.fixture
def lookup_mount_response():
    return {
        "request_id": "7a49be19-199b-ce19-c139-0c334bf07d72",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": {
            "accessor": "auth_approle_cafebabe",
            "config": {
                "allowed_response_headers": [""],
                "audit_non_hmac_request_keys": [""],
                "audit_non_hmac_response_keys": [""],
                "default_lease_ttl": 0,
                "force_no_cache": False,
                "max_lease_ttl": 0,
                "passthrough_request_headers": [""],
                "token_type": "default-service",
            },
            "deprecation_status": "supported",
            "description": "",
            "external_entropy_access": False,
            "local": False,
            "options": None,
            "plugin_version": "",
            "running_plugin_version": "v1.13.1+builtin.vault",
            "running_sha256": "",
            "seal_wrap": False,
            "type": "approle",
            "uuid": "testuuid",
        },
        "warnings": None,
    }


@pytest.fixture
def client():
    yield Mock(spec=vclient.AuthenticatedVaultClient)


@pytest.fixture
def approle_api(client):
    yield vapi.AppRoleApi(client)


@pytest.fixture
def identity_api(client):
    yield vapi.IdentityApi(client)


def test_list_approles(approle_api, client):
    """
    Ensure list_approles call the API as expected and returns only a list of names
    """
    client.list.return_value = {"data": {"keys": ["foo", "bar"]}}
    res = approle_api.list_approles(mount="salt-minions")
    assert res == ["foo", "bar"]
    client.list.assert_called_once_with("auth/salt-minions/role")


def test_destroy_secret_id_by_secret_id(approle_api, client):
    """
    Ensure destroy_secret_id calls the API as expected.
    """
    approle_api.destroy_secret_id(
        "test-minion", secret_id="test-secret-id", mount="salt-minions"
    )
    client.post.assert_called_once_with(
        "auth/salt-minions/role/test-minion/secret-id/destroy",
        payload={"secret_id": "test-secret-id"},
    )


def test_destroy_secret_id_by_accessor(approle_api, client):
    """
    Ensure destroy_secret_id calls the API as expected.
    """
    approle_api.destroy_secret_id(
        "test-minion", accessor="test-accessor", mount="salt-minions"
    )
    client.post.assert_called_once_with(
        "auth/salt-minions/role/test-minion/secret-id-accessor/destroy",
        payload={"secret_id_accessor": "test-accessor"},
    )


@pytest.mark.parametrize(
    "aliases",
    [
        [],
        [
            {"mount_accessor": "test-accessor", "id": "test-entity-alias-id"},
            {"mount_accessor": "other-accessor", "id": "other-entity-alias-id"},
        ],
    ],
)
def test_write_entity_alias(client, aliases, entity_fetch_response, identity_api):
    """
    Ensure write_entity_alias calls the API as expected.
    """
    metadata = {"foo": "bar"}
    payload = {
        "canonical_id": "test-entity-id",
        "mount_accessor": "test-accessor",
        "name": "test-role-id",
        "custom_metadata": metadata,
    }
    if aliases:
        entity_fetch_response["data"]["aliases"] = aliases
        if aliases[0]["mount_accessor"] == "test-accessor":
            payload["id"] = aliases[0]["id"]

    with patch(
        "salt.utils.vault.api.IdentityApi._lookup_mount_accessor",
        return_value="test-accessor",
    ), patch(
        "salt.utils.vault.api.IdentityApi.read_entity",
        return_value=entity_fetch_response["data"],
    ):
        identity_api.write_entity_alias(
            "salt_minion_test-minion",
            alias_name="test-role-id",
            mount="salt-minions",
            custom_metadata=metadata,
        )
        client.post.assert_called_with("identity/entity-alias", payload=payload)


def test_write_entity(client, identity_api):
    """
    Ensure write_entity calls the API as expected.
    """
    metadata = {"foo": "bar"}
    identity_api.write_entity("salt_minion_test-minion", metadata=metadata)
    payload = {"metadata": metadata}
    client.post.assert_called_with(
        "identity/entity/name/salt_minion_test-minion", payload=payload
    )


def test_read_entity_by_alias_failed(client, identity_api):
    """
    Ensure read_entity_by_alias raises VaultNotFoundError if the lookup fails.
    """
    with patch(
        "salt.utils.vault.api.IdentityApi._lookup_mount_accessor",
        return_value="test-accessor",
    ):
        client.post.return_value = []
        with pytest.raises(vapi.VaultNotFoundError):
            identity_api.read_entity_by_alias(
                alias="test-role-id", mount="salt-minions"
            )


def test_read_entity_by_alias(client, entity_lookup_response, identity_api):
    """
    Ensure read_entity_by_alias calls the API as expected.
    """
    with patch(
        "salt.utils.vault.api.IdentityApi._lookup_mount_accessor",
        return_value="test-accessor",
    ):
        client.post.return_value = entity_lookup_response
        res = identity_api.read_entity_by_alias(
            alias="test-role-id", mount="salt-minions"
        )
        assert res == entity_lookup_response["data"]
        payload = {
            "alias_name": "test-role-id",
            "alias_mount_accessor": "test-accessor",
        }
        client.post.assert_called_once_with("identity/lookup/entity", payload=payload)


def test_lookup_mount_accessor(client, identity_api, lookup_mount_response):
    """
    Ensure _lookup_mount_accessor calls the API as expected.
    """
    client.get.return_value = lookup_mount_response
    res = identity_api._lookup_mount_accessor("salt-minions")
    client.get.assert_called_once_with("sys/auth/salt-minions")
    assert res == "auth_approle_cafebabe"


@pytest.mark.parametrize("wrap", ["30s", False])
def test_generate_secret_id(
    client, wrapped_response, secret_id_response, wrap, approle_api
):
    """
    Ensure generate_secret_id calls the API as expected.
    """

    def res_or_wrap(*args, **kwargs):
        if kwargs.get("wrap"):
            return vaultutil.VaultWrappedResponse(**wrapped_response["wrap_info"])
        return secret_id_response

    client.post.side_effect = res_or_wrap
    metadata = {"foo": "bar"}
    res = approle_api.generate_secret_id(
        "test-minion", mount="salt-minions", metadata=metadata, wrap=wrap
    )
    if wrap:
        assert res == vaultutil.VaultWrappedResponse(**wrapped_response["wrap_info"])
    else:
        assert res == vaultutil.VaultSecretId(**secret_id_response["data"])
    client.post.assert_called_once_with(
        "auth/salt-minions/role/test-minion/secret-id",
        payload={"metadata": '{"foo": "bar"}'},
        wrap=wrap,
    )


@pytest.mark.parametrize("wrap", ["30s", False])
def test_read_role_id(client, wrapped_response, wrap, approle_api):
    """
    Ensure read_role_id calls the API as expected.
    """

    def res_or_wrap(*args, **kwargs):
        if kwargs.get("wrap"):
            return vaultutil.VaultWrappedResponse(**wrapped_response["wrap_info"])
        return {"data": {"role_id": "test-role-id"}}

    client.get.side_effect = res_or_wrap
    res = approle_api.read_role_id("test-minion", mount="salt-minions", wrap=wrap)
    if wrap:
        assert res == vaultutil.VaultWrappedResponse(**wrapped_response["wrap_info"])
    else:
        assert res == "test-role-id"
    client.get.assert_called_once_with(
        "auth/salt-minions/role/test-minion/role-id", wrap=wrap
    )


def test_read_approle(client, approle_api, approle_meta):
    """
    Ensure read_approle calls the API as expected.
    """
    client.get.return_value = {"data": approle_meta}
    res = approle_api.read_approle("test-minion", mount="salt-minions")
    assert res == approle_meta
    client.get.assert_called_once_with("auth/salt-minions/role/test-minion")


def test_write_approle(approle_api, client):
    """
    Ensure _manage_approle calls the API as expected.
    """
    policies = {"foo": "bar"}
    payload = {
        "token_explicit_max_ttl": 9999999999,
        "token_num_uses": 1,
        "token_policies": policies,
    }
    approle_api.write_approle("test-minion", mount="salt-minions", **payload)
    client.post.assert_called_once_with(
        "auth/salt-minions/role/test-minion", payload=payload
    )



"""
CloudStack Cloud Module
=======================

The CloudStack cloud module is used to control access to a CloudStack based
Public Cloud.

:depends: libcloud >= 0.15

Use of this module requires the ``apikey``, ``secretkey``, ``host`` and
``path`` parameters.

.. code-block:: yaml

    my-cloudstack-cloud-config:
      apikey: <your api key >
      secretkey: <your secret key >
      host: localhost
      path: /client/api
      driver: cloudstack

"""

# pylint: disable=function-redefined

import logging
import pprint

import salt.config as config
import salt.utils.cloud
import salt.utils.event
from salt.cloud.libcloudfuncs import *  # pylint: disable=redefined-builtin,wildcard-import,unused-wildcard-import
from salt.exceptions import SaltCloudSystemExit
from salt.utils.functools import namespaced_function
from salt.utils.versions import Version

# CloudStackNetwork will be needed during creation of a new node
# pylint: disable=import-error
try:
    from libcloud.compute.drivers.cloudstack import CloudStackNetwork

    # This work-around for Issue #32743 is no longer needed for libcloud >=
    # 1.4.0. However, older versions of libcloud must still be supported with
    # this work-around. This work-around can be removed when the required
    # minimum version of libcloud is 2.0.0 (See PR #40837 - which is
    # implemented in Salt 2018.3.0).
    if Version(libcloud.__version__) < Version("1.4.0"):
        # See https://github.com/saltstack/salt/issues/32743
        import libcloud.security

        libcloud.security.CA_CERTS_PATH.append("/etc/ssl/certs/YaST-CA.pem")
    HAS_LIBS = True
except ImportError:
    HAS_LIBS = False

# Get logging started
log = logging.getLogger(__name__)

# Redirect CloudStack functions to this module namespace
get_node = namespaced_function(get_node, globals())
get_size = namespaced_function(get_size, globals())
get_image = namespaced_function(get_image, globals())
avail_locations = namespaced_function(avail_locations, globals())
avail_images = namespaced_function(avail_images, globals())
avail_sizes = namespaced_function(avail_sizes, globals())
script = namespaced_function(script, globals())
list_nodes = namespaced_function(list_nodes, globals())
list_nodes_full = namespaced_function(list_nodes_full, globals())
list_nodes_select = namespaced_function(list_nodes_select, globals())
show_instance = namespaced_function(show_instance, globals())

__virtualname__ = "cloudstack"


# Only load in this module if the CLOUDSTACK configurations are in place
def __virtual__():
    """
    Set up the libcloud functions and check for CloudStack configurations.
    """
    if get_configured_provider() is False:
        return False

    if get_dependencies() is False:
        return False

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__,
        _get_active_provider_name() or __virtualname__,
        ("apikey", "secretkey", "host", "path"),
    )


def get_dependencies():
    """
    Warn if dependencies aren't met.
    """
    return config.check_driver_dependencies(__virtualname__, {"libcloud": HAS_LIBS})


def get_conn():
    """
    Return a conn object for the passed VM data
    """
    driver = get_driver(Provider.CLOUDSTACK)

    verify_ssl_cert = config.get_cloud_config_value(
        "verify_ssl_cert",
        get_configured_provider(),
        __opts__,
        default=True,
        search_global=False,
    )

    if verify_ssl_cert is False:
        try:
            import libcloud.security

            libcloud.security.VERIFY_SSL_CERT = False
        except (ImportError, AttributeError):
            raise SaltCloudSystemExit(
                "Could not disable SSL certificate verification. Not loading module."
            )

    return driver(
        key=config.get_cloud_config_value(
            "apikey", get_configured_provider(), __opts__, search_global=False
        ),
        secret=config.get_cloud_config_value(
            "secretkey", get_configured_provider(), __opts__, search_global=False
        ),
        secure=config.get_cloud_config_value(
            "secure",
            get_configured_provider(),
            __opts__,
            default=True,
            search_global=False,
        ),
        host=config.get_cloud_config_value(
            "host", get_configured_provider(), __opts__, search_global=False
        ),
        path=config.get_cloud_config_value(
            "path", get_configured_provider(), __opts__, search_global=False
        ),
        port=config.get_cloud_config_value(
            "port",
            get_configured_provider(),
            __opts__,
            default=None,
            search_global=False,
        ),
    )


def get_location(conn, vm_):
    """
    Return the node location to use
    """
    locations = conn.list_locations()
    # Default to Dallas if not otherwise set
    loc = config.get_cloud_config_value("location", vm_, __opts__, default=2)
    for location in locations:
        if str(loc) in (str(location.id), str(location.name)):
            return location


def get_security_groups(conn, vm_):
    """
    Return a list of security groups to use, defaulting to ['default']
    """
    securitygroup_enabled = config.get_cloud_config_value(
        "securitygroup_enabled", vm_, __opts__, default=True
    )
    if securitygroup_enabled:
        return config.get_cloud_config_value(
            "securitygroup", vm_, __opts__, default=["default"]
        )
    else:
        return False


def get_password(vm_):
    """
    Return the password to use
    """
    return config.get_cloud_config_value(
        "password",
        vm_,
        __opts__,
        default=config.get_cloud_config_value(
            "passwd", vm_, __opts__, search_global=False
        ),
        search_global=False,
    )


def get_key():
    """
    Returns the ssh private key for VM access
    """
    return config.get_cloud_config_value(
        "private_key", get_configured_provider(), __opts__, search_global=False
    )


def get_keypair(vm_):
    """
    Return the keypair to use
    """
    keypair = config.get_cloud_config_value("keypair", vm_, __opts__)

    if keypair:
        return keypair
    else:
        return False


def get_ip(data):
    """
    Return the IP address of the VM
    If the VM has  public IP as defined by libcloud module then use it
    Otherwise try to extract the private IP and use that one.
    """
    try:
        ip = data.public_ips[0]
    except Exception:  # pylint: disable=broad-except
        ip = data.private_ips[0]
    return ip


def get_networkid(vm_):
    """
    Return the networkid to use, only valid for Advanced Zone
    """
    networkid = config.get_cloud_config_value("networkid", vm_, __opts__)

    if networkid is not None:
        return networkid
    else:
        return False


def get_project(conn, vm_):
    """
    Return the project to use.
    """
    try:
        projects = conn.ex_list_projects()
    except AttributeError:
        # with versions <0.15 of libcloud this is causing an AttributeError.
        log.warning(
            "Cannot get projects, you may need to update libcloud to 0.15 or later"
        )
        return False
    projid = config.get_cloud_config_value("projectid", vm_, __opts__)

    if not projid:
        return False

    for project in projects:
        if str(projid) in (str(project.id), str(project.name)):
            return project

    log.warning("Couldn't find project %s in projects", projid)
    return False


def create(vm_):
    """
    Create a single VM from a data dict
    """
    try:
        # Check for required profile parameters before sending any API calls.
        if (
            vm_["profile"]
            and config.is_profile_configured(
                __opts__,
                _get_active_provider_name() or "cloudstack",
                vm_["profile"],
                vm_=vm_,
            )
            is False
        ):
            return False
    except AttributeError:
        pass

    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        "salt/cloud/{}/creating".format(vm_["name"]),
        sock_dir=__opts__["sock_dir"],
        args=__utils__["cloud.filter_event"](
            "creating", vm_, ["name", "profile", "provider", "driver"]
        ),
        transport=__opts__["transport"],
    )

    log.info("Creating Cloud VM %s", vm_["name"])
    conn = get_conn()
    # pylint: disable=not-callable
    kwargs = {
        "name": vm_["name"],
        "image": get_image(conn, vm_),
        "size": get_size(conn, vm_),
        "location": get_location(conn, vm_),
    }
    # pylint: enable=not-callable

    sg = get_security_groups(conn, vm_)
    if sg is not False:
        kwargs["ex_security_groups"] = sg

    if get_keypair(vm_) is not False:
        kwargs["ex_keyname"] = get_keypair(vm_)

    if get_networkid(vm_) is not False:
        kwargs["networkids"] = get_networkid(vm_)
        kwargs["networks"] = (  # The only attr that is used is 'id'.
            CloudStackNetwork(None, None, None, kwargs["networkids"], None, None),
        )

    if get_project(conn, vm_) is not False:
        kwargs["project"] = get_project(conn, vm_)

    event_data = kwargs.copy()
    event_data["image"] = kwargs["image"].name
    event_data["size"] = kwargs["size"].name

    __utils__["cloud.fire_event"](
        "event",
        "requesting instance",
        "salt/cloud/{}/requesting".format(vm_["name"]),
        sock_dir=__opts__["sock_dir"],
        args={
            "kwargs": __utils__["cloud.filter_event"](
                "requesting",
                event_data,
                ["name", "profile", "provider", "driver", "image", "size"],
            ),
        },
        transport=__opts__["transport"],
    )

    displayname = cloudstack_displayname(vm_)
    if displayname:
        kwargs["ex_displayname"] = displayname
    else:
        kwargs["ex_displayname"] = kwargs["name"]

    volumes = {}
    ex_blockdevicemappings = block_device_mappings(vm_)
    if ex_blockdevicemappings:
        for ex_blockdevicemapping in ex_blockdevicemappings:
            if "VirtualName" not in ex_blockdevicemapping:
                ex_blockdevicemapping["VirtualName"] = "{}-{}".format(
                    vm_["name"], len(volumes)
                )
            __utils__["cloud.fire_event"](
                "event",
                "requesting volume",
                "salt/cloud/{}/requesting".format(ex_blockdevicemapping["VirtualName"]),
                sock_dir=__opts__["sock_dir"],
                args={
                    "kwargs": {
                        "name": ex_blockdevicemapping["VirtualName"],
                        "device": ex_blockdevicemapping["DeviceName"],
                        "size": ex_blockdevicemapping["VolumeSize"],
                    }
                },
            )
            try:
                volumes[ex_blockdevicemapping["DeviceName"]] = conn.create_volume(
                    ex_blockdevicemapping["VolumeSize"],
                    ex_blockdevicemapping["VirtualName"],
                )
            except Exception as exc:  # pylint: disable=broad-except
                log.error(
                    "Error creating volume %s on CLOUDSTACK\n\n"
                    "The following exception was thrown by libcloud when trying to "
                    "requesting a volume: \n%s",
                    ex_blockdevicemapping["VirtualName"],
                    exc,
                    # Show the traceback if the debug logging level is enabled
                    exc_info_on_loglevel=logging.DEBUG,
                )
                return False
    else:
        ex_blockdevicemapping = {}
    try:
        data = conn.create_node(**kwargs)
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            "Error creating %s on CLOUDSTACK\n\n"
            "The following exception was thrown by libcloud when trying to "
            "run the initial deployment: \n%s",
            vm_["name"],
            exc,
            # Show the traceback if the debug logging level is enabled
            exc_info_on_loglevel=logging.DEBUG,
        )
        return False

    for device_name in volumes:
        try:
            conn.attach_volume(data, volumes[device_name], device_name)
        except Exception as exc:  # pylint: disable=broad-except
            log.error(
                "Error attaching volume %s on CLOUDSTACK\n\n"
                "The following exception was thrown by libcloud when trying to "
                "attach a volume: \n%s",
                ex_blockdevicemapping.get("VirtualName", "UNKNOWN"),
                exc,
                # Show the traceback if the debug logging level is enabled
                exc_info=log.isEnabledFor(logging.DEBUG),
            )
            return False

    ssh_username = config.get_cloud_config_value(
        "ssh_username", vm_, __opts__, default="root"
    )

    vm_["ssh_host"] = get_ip(data)
    vm_["password"] = data.extra["password"]
    vm_["key_filename"] = get_key()
    ret = __utils__["cloud.bootstrap"](vm_, __opts__)

    ret.update(data.__dict__)

    if "password" in data.extra:
        del data.extra["password"]

    log.info("Created Cloud VM '%s'", vm_["name"])
    log.debug(
        "'%s' VM creation details:\n%s", vm_["name"], pprint.pformat(data.__dict__)
    )

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{}/created".format(vm_["name"]),
        sock_dir=__opts__["sock_dir"],
        args=__utils__["cloud.filter_event"](
            "created", vm_, ["name", "profile", "provider", "driver"]
        ),
        transport=__opts__["transport"],
    )

    return ret


def destroy(name, conn=None, call=None):
    """
    Delete a single VM, and all of its volumes
    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )

    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        f"salt/cloud/{name}/destroying",
        sock_dir=__opts__["sock_dir"],
        args={"name": name},
    )

    if not conn:
        conn = get_conn()  # pylint: disable=E0602

    node = get_node(conn, name)  # pylint: disable=not-callable
    if node is None:
        log.error("Unable to find the VM %s", name)
    volumes = conn.list_volumes(node)
    if volumes is None:
        log.error("Unable to find volumes of the VM %s", name)
    # TODO add an option like 'delete_sshkeys' below
    for volume in volumes:
        if volume.extra["volume_type"] != "DATADISK":
            log.info(
                "Ignoring volume type %s: %s", volume.extra["volume_type"], volume.name
            )
            continue
        log.info("Detaching volume: %s", volume.name)
        __utils__["cloud.fire_event"](
            "event",
            "detaching volume",
            f"salt/cloud/{volume.name}/detaching",
            sock_dir=__opts__["sock_dir"],
            args={"name": volume.name},
        )
        if not conn.detach_volume(volume):
            log.error("Failed to Detach volume: %s", volume.name)
            return False
        log.info("Detached volume: %s", volume.name)
        __utils__["cloud.fire_event"](
            "event",
            "detached volume",
            f"salt/cloud/{volume.name}/detached",
            sock_dir=__opts__["sock_dir"],
            args={"name": volume.name},
        )

        log.info("Destroying volume: %s", volume.name)
        __utils__["cloud.fire_event"](
            "event",
            "destroying volume",
            f"salt/cloud/{volume.name}/destroying",
            sock_dir=__opts__["sock_dir"],
            args={"name": volume.name},
        )
        if not conn.destroy_volume(volume):
            log.error("Failed to Destroy volume: %s", volume.name)
            return False
        log.info("Destroyed volume: %s", volume.name)
        __utils__["cloud.fire_event"](
            "event",
            "destroyed volume",
            f"salt/cloud/{volume.name}/destroyed",
            sock_dir=__opts__["sock_dir"],
            args={"name": volume.name},
        )
    log.info("Destroying VM: %s", name)
    ret = conn.destroy_node(node)
    if not ret:
        log.error("Failed to Destroy VM: %s", name)
        return False
    log.info("Destroyed VM: %s", name)
    # Fire destroy action
    event = salt.utils.event.SaltEvent("master", __opts__["sock_dir"])
    __utils__["cloud.fire_event"](
        "event",
        "destroyed instance",
        f"salt/cloud/{name}/destroyed",
        sock_dir=__opts__["sock_dir"],
        args={"name": name},
    )
    if __opts__["delete_sshkeys"] is True:
        salt.utils.cloud.remove_sshkey(node.public_ips[0])
    return True


def block_device_mappings(vm_):
    """
    Return the block device mapping:

    ::

        [{'DeviceName': '/dev/sdb', 'VirtualName': 'ephemeral0'},
          {'DeviceName': '/dev/sdc', 'VirtualName': 'ephemeral1'}]
    """
    return config.get_cloud_config_value(
        "block_device_mappings", vm_, __opts__, search_global=True
    )


def cloudstack_displayname(vm_):
    """
    Return display name of VM:

    ::
        "minion1"
    """
    return config.get_cloud_config_value(
        "cloudstack_displayname", vm_, __opts__, search_global=True
    )



"""
Namecheap Nameserver Management

.. versionadded:: 2017.7.0

Prerequisites
-------------

This module uses the ``requests`` Python module to communicate to the namecheap
API.

Configuration
-------------

The Namecheap username, API key and URL should be set in the minion configuration
file, or in the Pillar data.

.. code-block:: yaml

    namecheap.name: companyname
    namecheap.key: a1b2c3d4e5f67a8b9c0d1e2f3
    namecheap.client_ip: 162.155.30.172
    #Real url
    namecheap.url: https://api.namecheap.com/xml.response
    #Sandbox url
    #namecheap.url: https://api.sandbox.namecheap.xml.response
"""

CAN_USE_NAMECHEAP = True


try:
    import salt.utils.namecheap
except ImportError:
    CAN_USE_NAMECHEAP = False


__virtualname__ = "namecheap_domains_ns"


def __virtual__():
    """
    Check to make sure requests and xml are installed and requests
    """
    if CAN_USE_NAMECHEAP:
        return "namecheap_domains_ns"
    return False


def get_info(sld, tld, nameserver):
    """
    Retrieves information about a registered nameserver. Returns the following
    information:

    - IP Address set for the nameserver
    - Domain name which was queried
    - A list of nameservers and their statuses

    sld
        SLD of the domain name

    tld
        TLD of the domain name

    nameserver
        Nameserver to retrieve

    CLI Example:

    .. code-block:: bash

        salt '*' namecheap_domains_ns.get_info sld tld nameserver
    """
    opts = salt.utils.namecheap.get_opts("namecheap.domains.ns.delete")
    opts["SLD"] = sld
    opts["TLD"] = tld
    opts["Nameserver"] = nameserver

    response_xml = salt.utils.namecheap.post_request(opts)
    if response_xml is None:
        return {}

    domainnsinforesult = response_xml.getElementsByTagName("DomainNSInfoResult")[0]

    return salt.utils.namecheap.xml_to_dict(domainnsinforesult)


def update(sld, tld, nameserver, old_ip, new_ip):
    """
    Deletes a nameserver. Returns ``True`` if the nameserver was updated
    successfully.

    sld
        SLD of the domain name

    tld
        TLD of the domain name

    nameserver
        Nameserver to create

    old_ip
        Current ip address

    new_ip
        New ip address

    CLI Example:

    .. code-block:: bash

        salt '*' namecheap_domains_ns.update sld tld nameserver old_ip new_ip
    """
    opts = salt.utils.namecheap.get_opts("namecheap.domains.ns.update")
    opts["SLD"] = sld
    opts["TLD"] = tld
    opts["Nameserver"] = nameserver
    opts["OldIP"] = old_ip
    opts["IP"] = new_ip

    response_xml = salt.utils.namecheap.post_request(opts)
    if response_xml is None:
        return False

    domainnsupdateresult = response_xml.getElementsByTagName("DomainNSUpdateResult")[0]
    return salt.utils.namecheap.string_to_value(
        domainnsupdateresult.getAttribute("IsSuccess")
    )


def delete(sld, tld, nameserver):
    """
    Deletes a nameserver. Returns ``True`` if the nameserver was deleted
    successfully

    sld
        SLD of the domain name

    tld
        TLD of the domain name

    nameserver
        Nameserver to delete

    CLI Example:

    .. code-block:: bash

        salt '*' namecheap_domains_ns.delete sld tld nameserver
    """
    opts = salt.utils.namecheap.get_opts("namecheap.domains.ns.delete")
    opts["SLD"] = sld
    opts["TLD"] = tld
    opts["Nameserver"] = nameserver

    response_xml = salt.utils.namecheap.post_request(opts)
    if response_xml is None:
        return False

    domainnsdeleteresult = response_xml.getElementsByTagName("DomainNSDeleteResult")[0]
    return salt.utils.namecheap.string_to_value(
        domainnsdeleteresult.getAttribute("IsSuccess")
    )


def create(sld, tld, nameserver, ip):
    """
    Creates a new nameserver. Returns ``True`` if the nameserver was created
    successfully.

    sld
        SLD of the domain name

    tld
        TLD of the domain name

    nameserver
        Nameserver to create

    ip
        Nameserver IP address

    CLI Example:

    .. code-block:: bash

        salt '*' namecheap_domains_ns.create sld tld nameserver ip
    """

    opts = salt.utils.namecheap.get_opts("namecheap.domains.ns.create")
    opts["SLD"] = sld
    opts["TLD"] = tld
    opts["Nameserver"] = nameserver
    opts["IP"] = ip

    response_xml = salt.utils.namecheap.post_request(opts)
    if response_xml is None:
        return False

    domainnscreateresult = response_xml.getElementsByTagName("DomainNSCreateResult")[0]
    return salt.utils.namecheap.string_to_value(
        domainnscreateresult.getAttribute("IsSuccess")
    )
